#include "paint.h"
#include <keyboard.h>
#include <mouse.h>
#include <thread.h>

Image *canvasimg;
Memimage *canvas;
Memimage *canvasmask;
uchar *canvastmp;
int canvaslen;

/* relative to screen.min */
Point canvaspos;
Point patternpalettepos;
Rectangle patternpaletterect;
Point colorpalettepos;
Rectangle colorpaletterect;

Point swatchdim;

Memimage **patterns;
int currentpattern;

ulong fgcolor;

int colors[] = {
	0x0000007F,
	0xFFFFFF7F,
	0x68372B7F,
	0x70A4B27F,
	0x6F3D867F,
	0x588D437F,
	0x3528797F,
	0xB8C76F7F,
	0x6F4F257F,
	0x4339007F,
	0x9A67597F,
	0x4444447F,
	0x6C6C6C7F,
	0x9AD2847F,
	0x6C5EB57F,
	0x9595957F,
};

void *
emalloc(ulong n)
{
	void *p;
	if((p = malloc(n)) == nil)
		sysfatal("no memory");
	return p;
}

void
kbdproc(void *arg)
{
	Keyboardctl *kctl = arg;
	Rune r;

	for(;;) {
		recv(kctl->c, &r);
		switch(r) {
		case Kdel:
			threadexitsall(nil);
		}
	}
}

void
redrawcanvas(Rectangle dirty, int flush)
{
	Rectangle r;

	r = rectaddpt(rectaddpt(dirty, screen->r.min), canvaspos);
	gendraw(screen, r, canvasimg, dirty.min, nil, ZP);
	if(flush)
		flushimage(display, 1);
}

void
drawshadow(Image *dst, Rectangle r)
{
	static Image *shadow;

	if(shadow == nil) {
		if((shadow = allocimage(display, Rect(0, 0, 1, 1), RGBA32, 1, 0x0000007F)) == nil)
			sysfatal("allocimage shadow: %r");
	}

	draw(dst, Rect(r.min.x+4, r.max.y+1, r.max.x+4, r.max.y+4), shadow, nil, ZP);
	draw(dst, Rect(r.max.x+1, r.min.y+4, r.max.x+4, r.max.y+1), shadow, nil, ZP);
}

Image *
mkpatternpalette(void)
{
	static Memimage *mask, *mem, *memfg;
	static int tmplen;
	static uchar *tmp;
	static Image *img;

	Rectangle swatchr;
	int i, n;

	if(mask == nil) {
		if((mask = allocmemimage(Rect(0, 0, swatchdim.x * npatterns, swatchdim.y), GREY8)) == nil)
			sysfatal("mkpatternpalette: alloc mask: %r");
		if((mem = allocmemimage(mask->r, RGB24)) == nil)
			sysfatal("mkpatternpalette: alloc mem: %r");
		tmp = emalloc(tmplen = Dy(mem->r) * Dx(mem->r) * 3);
		if((img = allocimage(display, mask->r, RGB24, 0, DNofill)) == nil)
			sysfatal("mkpatternpalette: alloc img: %r");

		if((memfg = allocmemimage(Rect(0, 0, 1, 1), RGBA32)) == nil)
			sysfatal("mkpatternpalette: alloc fg: %r");
		memfg->flags |= Frepl;
		memfg->clipr = INFR;

		for(i = 0; i < npatterns; i++) {
			swatchr = Rect(swatchdim.x * i, 0, swatchdim.x * (i+1), swatchdim.y);
			memimagedraw(mask, swatchr, patterns[i], ZP, nil, ZP, SoverD);
		}

		patternpaletterect = rectaddpt(mask->r, patternpalettepos);
	}

	memfillcolor(mem, 0x7F7F7FFF);
	memfillcolor(memfg, colors[fgcolor]);
	memimagedraw(mem, mem->r, memfg, ZP, mask, ZP, SoverD);
	if((n = unloadmemimage(mem, mem->r, tmp, tmplen)) < 0) {
		werrstr("unload: %r");
		return nil;
	}
	if(loadimage(img, img->r, tmp, n) < 0) {
		werrstr("load: %r");
		return nil;
	}
	return img;
}

Image *
mkcolorpalette(void)
{
	Memimage *mem, *c;
	int tmplen;
	uchar *tmp;
	Image *img;
	int i, n;
	Rectangle swatchr;

	if((mem = allocmemimage(Rect(0, 0, swatchdim.x * nelem(colors), swatchdim.y), RGB24)) == nil)
		sysfatal("mkcolorpalette: alloc mem: %r");
	tmp = emalloc(tmplen = Dy(mem->r) * Dx(mem->r) * 3);
	if((img = allocimage(display, mem->r, RGB24, 0, DNofill)) == nil)
		sysfatal("mkcolorpalette: alloc img: %r");

	if((c = allocmemimage(Rect(0, 0, 1, 1), RGB24)) == nil)
		sysfatal("mkcolorpalette: alloc c: %r");
	c->flags |= Frepl;
	c->clipr = INFR;

	for(i = 0; i < nelem(colors); i++) {
		swatchr = Rect(swatchdim.x * i, 0, swatchdim.x * (i+1), swatchdim.y);
		memfillcolor(c, colors[i]);
		memimagedraw(mem, swatchr, c, ZP, nil, ZP, SoverD);
	}

	if((n = unloadmemimage(mem, mem->r, tmp, tmplen)) < 0)
		sysfatal("mkcolorpalette: unload: %r");
	if(loadimage(img, img->r, tmp, n) < 0)
		sysfatal("mkcolorpalette: load: %r");

	colorpaletterect = rectaddpt(img->r, colorpalettepos);

	freememimage(mem);
	free(tmp);
	freememimage(c);
	return img;
}

Rectangle
swatchrect(Image *img, Point pos, int swatchidx)
{
	Point min, max;

	min = addpt(addpt(img->r.min, screen->r.min), pos);
	min.x += swatchdim.x * swatchidx;
	max = addpt(min, swatchdim);
	return Rpt(min, max);
}

void
redrawpatternpalette(int dirty)
{
	static Image *palette;

	if(palette == nil) {
		if((palette = mkpatternpalette()) == nil)
			sysfatal("mkpatternpalette: %r");
	} else if(dirty) {
		if((palette = mkpatternpalette()) == nil)
			fprint(2, "mkpatternpalette: %r\n");
	}

	draw(screen, rectaddpt(screen->r, patternpalettepos), palette, nil, ZP);
	border(screen, swatchrect(palette, patternpalettepos, currentpattern), 2, display->black, ZP);

	if(dirty)
		flushimage(display, 1);
}

void
redrawcolorpalette(int dirty)
{
	static Image *img;

	if(img == nil && (img = mkcolorpalette()) == nil)
		sysfatal("mkcolorpalette: %r");

	draw(screen, rectaddpt(screen->r, colorpalettepos), img, nil, ZP);
	border(screen, swatchrect(img, colorpalettepos, fgcolor), 2, display->black, ZP);

	if(dirty)
		flushimage(display, 1);
}


void
redraw(void)
{
	static int initialized;
	static Image *bg, *shadow, *patternpalette, *colorpalette;

	if(!initialized && (initialized = 1)) {
		if((bg = allocimage(display, Rect(0, 0, 8, 2), GREY8, 1, DNofill)) == nil)
			sysfatal("allocimage bg: %r");
		uchar bgbits[] = {
			0x00, 0x00, 0x00, 0x00, 0x8F, 0x8F, 0x8F, 0x8F,
			0x8F, 0x8F, 0x8F, 0x8F, 0x00, 0x00, 0x00, 0x00,
		};
		if(loadimage(bg, bg->r, bgbits, sizeof bgbits) < 0)
			sysfatal("loadimage bg: %r");

		if((patternpalette = mkpatternpalette()) == nil)
			sysfatal("mkpatternpalette: %r");
		if((colorpalette = mkcolorpalette()) == nil)
			sysfatal("mkcolorpalette: %r");
	}

	Rectangle r;

	r = rectaddpt(rectaddpt(canvasimg->r, screen->r.min), canvaspos);
	draw(screen, screen->r, bg, nil, ZP);
	redrawcanvas(canvasimg->r, 0);
	border(screen, r, -1, display->black, ZP);
	drawshadow(screen, r);

	redrawpatternpalette(0);
	r = rectaddpt(rectaddpt(patternpalette->r, screen->r.min), patternpalettepos);
	border(screen, r, -1, display->black, ZP);
	drawshadow(screen, r);

	redrawcolorpalette(0);
	r = rectaddpt(rectaddpt(colorpalette->r, screen->r.min), colorpalettepos);
	border(screen, r, -1, display->black, ZP);
	drawshadow(screen, r);

	flushimage(display, 1);
}

int
loadcanvas(Rectangle r)
{
	int n;

	if((n = unloadmemimage(canvas, r, canvastmp, canvaslen)) < 0) {
		werrstr("unload canvas");
		return -1;
	}

	if(loadimage(canvasimg, r, canvastmp, n) < 0) {
		werrstr("load canvasimg");
		return -1;
	}

	return 0;
}

void
brush(Mousectl *mctl)
{
	static Memimage *c;
	static int radius = 15;
	Memimage *pattern;
	Point lastpos, m;
	Rectangle dirty;

	if(mctl->buttons & 8) {
		radius++;
		return;
	} else if(mctl->buttons & 16) {
		radius = radius > 0 ? radius - 1 : 0;
		return;
	}

	if((mctl->buttons & 1) == 0)
		return;

	if(c == nil) {
		if((c = allocmemimage(Rect(0, 0, 1, 1), RGBA32)) == nil) {
			fprint(2, "brush: alloc c: %r\n");
			return;
		}
		c->flags |= Frepl;
		c->clipr = INFR;
	}
	memfillcolor(c, colors[fgcolor]);

	pattern = patterns[currentpattern];
	lastpos = subpt(subpt(mctl->xy, screen->r.min), canvaspos);
	do {
		m = subpt(subpt(mctl->xy, screen->r.min), canvaspos);
		dirty = memlinebbox(lastpos, m, 1, 1, radius);
		if(rectXrect(dirty, canvasimg->r)) {
			rectclip(&dirty, canvasimg->r);
			memimageline(canvasmask, lastpos, m, 1, 1, radius, memopaque, ZP, SoverD);
			memimagedraw(canvasmask, dirty, pattern, dirty.min, canvasmask, dirty.min, SoverD);
			memimagedraw(canvas, dirty, c, dirty.min, canvasmask, dirty.min, SoverD);
			memimagedraw(canvasmask, dirty, memtransparent, dirty.min, nil, ZP, Clear);
			if(loadcanvas(dirty) < 0)
				fprint(2, "brush: load: %r: %R\n", dirty);
			redrawcanvas(dirty, 1);
		}
		lastpos = m;
		readmouse(mctl);
	} while(mctl->buttons & 1);
}

void
pickpattern(Mousectl *mctl)
{
	Point m;
	int newpattern;

	if(mctl->buttons & (8|16)) {
		if(mctl->buttons & 8 && currentpattern > 0)
			currentpattern--;
		else if(mctl->buttons & 16 && currentpattern < npatterns-1)
			currentpattern++;
		redrawpatternpalette(1);
		return;
	}

	do {
		m = subpt(subpt(mctl->xy, screen->r.min), patternpalettepos);
		newpattern = m.x / swatchdim.x;
		if(newpattern < 0)
			newpattern = 0;
		else if(newpattern > npatterns - 1)
			newpattern = npatterns - 1;
		currentpattern = newpattern;
		redrawpatternpalette(1);
		readmouse(mctl);
	} while(mctl->buttons);
}

void
pickcolor(Mousectl *mctl)
{
	Point m;
	int newcolor;

	if(mctl->buttons & (8|16)) {
		if(mctl->buttons & 8 && fgcolor > 0)
			fgcolor--;
		else if(mctl->buttons & 16 && fgcolor < nelem(colors)-1)
			fgcolor++;
		redrawcolorpalette(1);
		redrawpatternpalette(1);
		return;
	}

	do {
		m = subpt(subpt(mctl->xy, screen->r.min), colorpalettepos);
		newcolor = m.x / swatchdim.x;
		if(newcolor < 0)
			newcolor = 0;
		else if(newcolor > nelem(colors) - 1)
			newcolor = nelem(colors) - 1;
		fgcolor = newcolor;
		redrawcolorpalette(1);
		redrawpatternpalette(1);
		readmouse(mctl);
	} while(mctl->buttons);
}

void
mouseproc(void *arg)
{
	Mousectl *mctl = arg;

	for(;;) {
		readmouse(mctl);
		if(mctl->buttons && ptinrect(subpt(mctl->xy, screen->r.min), patternpaletterect))
			pickpattern(mctl);
		else if(mctl->buttons && ptinrect(subpt(subpt(mctl->xy, screen->r.min), canvaspos), canvas->r))
			brush(mctl);
		else if(mctl->buttons && ptinrect(subpt(mctl->xy, screen->r.min), colorpaletterect))
			pickcolor(mctl);
	}
}

void
resizeproc(void *arg)
{
	Mousectl *mctl = arg;
	for(;;) {
		recvul(mctl->resizec);
		lockdisplay(display);
		if(getwindow(display, Refnone) < 0)
			sysfatal("getwindow: %r");
		redraw();
		unlockdisplay(display);
	}
}

void
initpatterns(void)
{
	Memimage **p;
	int i;
	Protopattern *pp;

	patterns = emalloc(npatterns*sizeof(Memimage));
	p = patterns;

	for(i = 0; i < npatterns; i++) {
		pp = ppatterns[i];
		if((*p = allocmemimage(Rect(0, 0, pp->Δx, pp->Δy), pp->chan)) == nil)
			sysfatal("alloc pattern: %r");
		(*p)->flags |= Frepl;
		(*p)->clipr = INFR;
		if(loadmemimage(*p, (*p)->r, pp->data, pp->ndata) < 0)
			sysfatal("load pattern: %r");
		p++;
	}
}

void
usage(void)
{
	fprint(2, "usage: %s\n", argv0);
	threadexitsall("usage");
}

void
threadmain(int argc, char **argv)
{
	Keyboardctl *kctl;
	Mousectl *mctl;

	ARGBEGIN {
	default:
		usage();
	} ARGEND

	if(initdraw(nil, nil, argv0) < 0)
		sysfatal("initdraw: %r");
	display->locking = 1;
	unlockdisplay(display);
	memimageinit();
	if((kctl = initkeyboard(nil)) == nil)
		sysfatal("initkeyboard: %r");
	if((mctl = initmouse(nil, screen)) == nil)
		sysfatal("initmouse: %r");

	INFR = Rect(-0x3FFFFFF, -0x3FFFFFF, 0x3FFFFFF, 0x3FFFFFF);

	if((canvasimg = allocimage(display, Rect(0, 0, 480, 360), RGB24, 0, DWhite)) == nil)
		sysfatal("alloc canvasimg: %r");
	if((canvas = allocmemimage(canvasimg->r, canvasimg->chan)) == nil)
		sysfatal("alloc canvas: %r");
	if((canvasmask = allocmemimage(canvasimg->r, canvasimg->chan)) == nil)
		sysfatal("alloc canvasmask: %r");
	canvaslen = Dy(canvasimg->r) * Dx(canvasimg->r) * 3;
	canvastmp = emalloc(canvaslen);
	memfillcolor(canvas, 0x7F7F7FFF);
	memfillcolor(canvasmask, DTransparent);
	if(loadcanvas(canvas->r) < 0)
		sysfatal("loadcanvas: %r");
	canvaspos = Pt(80, 40);
	swatchdim = Pt(32, 32);
	patternpalettepos = Pt(canvaspos.x, canvaspos.y + Dy(canvas->r) + 32);
	colorpalettepos = Pt(patternpalettepos.x, patternpalettepos.y + swatchdim.y + 16);

	initpatterns();

	proccreate(kbdproc, kctl, 1024);
	proccreate(mouseproc, mctl, 8*1024);
	proccreate(resizeproc, mctl, 8*1024);

	redraw();
	for(;;)
		sleep(1000*1000);
}
